import json
from pprint import pformat

from stayfulrestsdk.api import get_default_api

class Resource(object):
    embed = None

    def __init__(self, attrs=None, api=None):
        self.__dict__['api'] = api or get_default_api()

        self.__dict__['attrs'] = {}
        self.merge(attrs or {})

    @property
    def url(self):
        ''' Can be either whole url or just path part '''
        url = None
        if 'url' in self.attrs:
            url = self.attrs['url']
        elif hasattr(self, 'get_url'):
            url = self.get_url()

        if url is None:
            raise APIException('Cannot retrieve or build resource url')

        return url

    def convert(self, name, value):
        ''' Converts value to Resource if it is in self.embed, 
            returns original value otherwise 
        '''

        # we need to convert convert only if embed is 
        # defined and it contains current attribute
        if not self.embed or name not in self.embed:
            return value
        
        cls = self.embed[name]

        # value can be a single resource, or a list 
        # of resources, or a dict with 'items' list
        if isinstance(value, (list, tuple)):
            # list of resources
            return [cls(item, api=self.api) for item in value]
        elif isinstance(value, dict):
            if isinstance(value.get('items'), (list, tuple)):
                # dict with 'items' element
                return [cls(item, api=self.api) for item in value['items']]
            else:
                # single resource
                return cls(value, api=self.api)

    def merge(self, attrs):
        # use setattr on each attribute to make
        # sure that .convert() is called
        for key, value in attrs.items():
            setattr(self, key, value)

    def __setattr__(self, name, value):
        if name in self.__dict__:
            super(Resource, self).__setattr__(name, value)
        else:
            self.attrs[name] = self.convert(name, value)

    def __getattr__(self, name):
        try:
            return self.attrs[name]
        except KeyError as e:
            raise AttributeError(*e.args)

    def __repr__(self):
        classname = type(self).__name__
        data = pformat(self.attrs, indent=4)
        return '<{classname}: {data}>'.format(classname=classname, data=data)


class DeleteMixin(Resource):
    def delete(self):
        self.api.request('delete', self.url)
        return True


class CreateMixin(Resource):
    def create(self):
        url = self.api.build_endpoint(self.path)
        resp = self.api.request('post', url, data=self.attrs)
        self.merge(resp.json())
        return self


class GetMixin(Resource):
    @classmethod
    def get(cls, resource_id, api=None):
        api = api or get_default_api()
        resource_url = '{}/{}'.format(cls.path, resource_id)
        resp = api.request('get', resource_url)
        return cls(resp.json())

    def refresh(self):
        resp = self.api.request('get', self.url)
        self.merge(resp.json())
        return self